const l0 = require("./l0")

const TYPE_MD = {
    raw:10,q:11,lrq:12,
    10:'raw',11:'q',12:'lrq'
}


class RawTok {
    constructor(chunk,l0_raw_tok) {
        this.type = 10;
        this.raw = chunk.slice(l0_raw_tok.si,l0_raw_tok.ei)
    }
    stringify() {return(this.raw)}
}


class Qtok {
    constructor(chunk,l0_qtok) {
        this.type = 11;
        this.raw = chunk.slice(l0_qtok.si,l0_qtok.ei)
    }
    stringify() {return(this.raw)}
}




function map_l0_toks(chunk,toks) {
    return(
        toks.map(
            l0tok => {
                if(l0.is_raw_tok(l0tok)) {
                    return(new RawTok(chunk,l0tok))
                } else {
                    return(new Qtok(chunk,l0tok))
                }
            }
        )
    )
}


/////
class LRQtok {
    static tl_to_lref(tl) {return(Object.fromEntries(tl))}
    static tl_to_rref(tl) {
        tl = tl.map(e=>[e[1],e[0]])
        return(Object.fromEntries(tl))
    }
    static tl_to_lrref(tl) {
        let lref = {}
        let rref = {}
        for(let e of tl) {
            lref[e[0]] = e[1];
            rref[e[1]] = e[0];
        }
        return([lref,rref])
    }
    static lref_to_tl(lref) {return(Object.entries(lref))}
    static rref_to_tl(rref) {
        let tl = Object.entries(rref);
        tl = tl.map(e=>[e[1],e[0]]);
        return(tl)
    }
    static from_tl(qtok,tl) {
        let q = qtok.raw;
        let [lref,rref] = tl_to_lrref(tl);
        if(lref.hasOwnProperty(q)) {
            return(new LRQtok(qtok,q,lref[q]))
        } else if(rref.hasOwnProperty(q)) {
            return(new LRQtok(qtok,rref[q],q))
        } else {
        }
    }
    static to_twin_ref(ref) {return(LRQtok.rref_to_tl(ref))}
    static from_lref(qtok,lref) {
        let tl = Object.entries(lref);
        return(LRQtok.from_tl(qtok,tl))
    }
    static from_rref(qtok,rref) {
        let tl = LRQtok.rref_to_tl(rref) 
        return(LRQtok.from_tl(qtok,tl))
    }
    ////
    constructor(qtok,lq,rq) {
        this.q = qtok.raw;
        this.lq = lq??q;
        this.rq = rq??this.lq;
        this.type = 12;
    }
    ////
    is_lq()     {return(this.q === this.lq)}
    is_rq()     {return(this.q === this.rq)}
    is_lrq()    {return(this.q === this.lq && this.q === this.rq)}
    ////
    get pair_() {return([this.lq,this.rq])}
    get twin_() {return(this.is_lq()?this.rq:this.lq)}
    ////
    matchr(tok) {return(this.rq === tok.q && this.is_lq() && tok.is_rq())}
    matchl(tok) {return(this.lq === tok.q && this.is_rq() && tok.is_lq())}
    match(tok)  {return(matchr(tok) || match(tok))}
    stringify() {return(this.q)}
}



/////



module.exports = {
    RawTok,
    Qtok,
    map_l0_toks,
    TYPE_MD,
    is_raw_tok:(o)=>(o instanceof RawTok),
    is_qtok:(o)=>(o instanceof Qtok),
    LRQtok,
    is_lrqtok:(o)=>(o instanceof LRQtok),
    is_l1_tok:(o)=> (o.type>=10 && o.type<100),
}
